; Animated Sprite
;         on Atari 2600
;         by Sebastian Mihai, http://sebastianmihai.com
;
;
; This is a demonstration program of how to implement an animated sprite
; which is controllable via the joystick.
;
; I've programmed games before, but writing for the Atari 2600 is unlike
; other consoles. By far the largest difficulty lies in the fact that the
; interesting parts of the program are an exercise in CPU timing.
;
; Here are some of the aspects new to me which had to be overcome:
;     - working with finely adjusted sprite X position
;     - "setting" sprite's onscreen X position by means of CPU timing
;     - tightening show/no show sprite code such that its vertical position
;       is controlled while remaining fast enough that it does not 
;       distort the sprite
;     - spreading out computation (checking input, recalculating sprite 
;       position) across vertically blanked (invisible) scanlines, to keep
;       the code which executes during visible scanlines as fast as possible
;     - when to set the background's colour so that the scanline is not
;       incompletely coloured

    PROCESSOR 6502
    include "atari2600\vcs.h"
    include "atari2600\macro.h"

; constants
SPRITE_HEIGHT = 10
SPRITE_WIDTH = 8
SPRITE_ANIMATION_FRAME_COUNT = 4
SPRITE_TOTAL_FRAME_LINES = SPRITE_ANIMATION_FRAME_COUNT * SPRITE_HEIGHT
BOTTOM_LAYER_HEIGHT = 20
SKY_COLOUR = $73

; variables, each value representing the address
; in RAM of the respective variable
SPRITE0_TOP_Y = $80
SPRITE0_BOTTOM_Y = $81
SPRITE0_LEFT_X = $82
SPRITE0_RIGHT_X = $83
SPRITE0_LEFT_X_COARSE = $84
SPRITE0_PIXEL_OFFSET = $85
CURRENT_SCANLINE = $86
FRAME_COUNTER = $87


    SEG
    ORG $F000

title_string
    dc.b "Animated Sprite by Sebastian Mihai "
    
; each entry is displayed as a horizontal line of 8 pixels
; each group of entries represents an animation frame
sprite_bitmap
    dc.b %00011000
    dc.b %00111100
    dc.b %00011000
    dc.b %10011000
    dc.b %01111110
    dc.b %00011001
    dc.b %00011000
    dc.b %00011000
    dc.b %00100100
    dc.b %01000010
    
    dc.b %00011000
    dc.b %00111100
    dc.b %00011000
    dc.b %00011000
    dc.b %11111111
    dc.b %00011000
    dc.b %00011000
    dc.b %00011000
    dc.b %00011000
    dc.b %00100100
    
    dc.b %00011000
    dc.b %00111100
    dc.b %00011000
    dc.b %00011001
    dc.b %01111110
    dc.b %10011000
    dc.b %00011000
    dc.b %00011000
    dc.b %00011000
    dc.b %00011000
    
    dc.b %00011000
    dc.b %00111100
    dc.b %00011000
    dc.b %00011000
    dc.b %11111111
    dc.b %00011000
    dc.b %00011000
    dc.b %00011000
    dc.b %00011000
    dc.b %00100100
    
; used to compute the horizontal fine adjustment, needed to 
; position a sprite more accurately
fine_adjust_table
    ; left
    dc.b %01110000  ; 7
    dc.b %01100000  ; 6
    dc.b %01010000  ; 5
    dc.b %01000000  ; 4
    dc.b %00110000  ; 3
    dc.b %00100000  ; 2
    dc.b %00010000  ; 1
    dc.b %00000000  ; 0
    ; right
    dc.b %11110000  ; -1
    dc.b %11100000  ; -2
    dc.b %11010000  ; -3
    dc.b %11000000  ; -4
    dc.b %10110000  ; -5
    dc.b %10100000  ; -6
    dc.b %10010000  ; -7
    dc.b %10000000  ; -8
    

; must be called after sprite's Y change
spr0_moved_vertically
    clc
    lda SPRITE0_TOP_Y
    adc SPRITE_HEIGHT
    sta SPRITE0_BOTTOM_Y
    rts
    
    
; must be called after sprite's X changes, to compute horizontal
; fine adjustment
;
; this is because positioning a sprite on the Atari 2600 involves two steps:
;     1. computing a coarse position
;     2. computing a fine adjustment
spr0_moved_horizontally
    clc
    lda SPRITE0_LEFT_X
    
    pha
    adc SPRITE_WIDTH
    sta SPRITE0_RIGHT_X
    pla                     ; A := sprite X
    
    lsr                         ;
    lsr                         ;
    lsr                         ;
    lsr                         ;
    sta SPRITE0_LEFT_X_COARSE   ; sprite X coarse := sprite X div 16
    
    ; apply horizontal fine adjustment
    lda SPRITE0_LEFT_X
    and #$0F
    tax                     ; X := (sprite0 x location) mod 16
    lda fine_adjust_table,x ; A := horizontaly fine adjustment
    sta HMP0                ; fine-adjust sprite 0 horizontally
    rts


start
    cld                     ; disable decimal mode
    ldx #$FF
    txs                     ; SP := 0xFF
    inx                     ; X := 0
    txa                     ; A := 0
    tay                     ; Y := 0

    lda #$0F
    sta COLUP0              ; set sprite colour

    lda #0                  ; initial sprite X
    sta SPRITE0_LEFT_X
    jsr spr0_moved_horizontally

    lda #1                  ; initial sprite Y
    sta SPRITE0_TOP_Y
    jsr spr0_moved_vertically

    lda #0
    sta SPRITE0_PIXEL_OFFSET
    sta FRAME_COUNTER
    
; this is reached at the beginning of each video frame
video_frame_start

;=============================================================================
; VBLANK
;=============================================================================
    lda #0
    sta VBLANK

    lda #2
    sta VSYNC
    
;=============================================================================
; 3 scanlines of VSYNC
;=============================================================================
    REPEAT 3
        sta WSYNC
    REPEND

    lda #0
    sta VSYNC

;=============================================================================
; 37 scanlines of vertical blank now follow
;=============================================================================       
    REPEAT 29
        sta WSYNC
    REPEND
    
    ; scanline 30 begins
    
    ; before vertical blank scanline we check
    ; some inputs
    lda SPRITE0_LEFT_X
    cmp #0
    beq after_checked_left_controller_left
    
    lda #%01000000          ; check left controller LEFT
	bit SWCHA               ; 
	bne after_checked_left_controller_left
	; left controller is held down
    dec SPRITE0_LEFT_X       ; move sprite left
after_checked_left_controller_left
    sta WSYNC               ; halt for scanline
    
    ; scanline 31 begins
    
    ; before vertical blank scanline we check
    ; some inputs
    lda SPRITE0_RIGHT_X
    cmp #160 + SPRITE_WIDTH - 3
    bcs after_checked_left_controller_right
    
    lda #%10000000          ; check left controller RIGHT
	bit SWCHA               ; 
	bne after_checked_left_controller_right
	; left controller is held down
    inc SPRITE0_LEFT_X       ; move sprite right
after_checked_left_controller_right
    sta WSYNC               ; halt for scanline
    
    ; scanline 32 begins
    
    ; before vertical blank scanline we check
    ; some inputs
    lda SPRITE0_TOP_Y
    cmp #1
    beq after_checked_left_controller_up
    
    lda #%00010000          ; check left controller UP
	bit SWCHA               ; 
	bne after_checked_left_controller_up
	; left controller is held down
    dec SPRITE0_TOP_Y       ; move sprite up
    jsr spr0_moved_vertically
after_checked_left_controller_up
    sta WSYNC               ; halt for scanline
    
    ; scanline 33 begins
    
    ; before the vertical blank scanline we check
    ; some inputs
    lda SPRITE0_BOTTOM_Y
    cmp #192 - BOTTOM_LAYER_HEIGHT - 1
    bcs after_checked_left_controller_down
    
    lda #%00100000          ; check left controller DOWN
	bit SWCHA               ; 
	bne after_checked_left_controller_down
	; left controller is held down
    inc SPRITE0_TOP_Y       ; move sprite down
    jsr spr0_moved_vertically
after_checked_left_controller_down
    sta WSYNC               ; halt for scanline
    
    ; scanline 34 begins
    
    jsr spr0_moved_horizontally ; doing this on the same scanline as
                                ; the reading of input and changing
                                ; sprite X location would put that
                                ; scanline over the limit, so we
                                ; do it in its own scanline
                                
    sta WSYNC               ; halt for scanline
    
    ; scanline 35 begins
    
    ; this scanline is very important because it is where we write to
    ; RESPx, to tell TIA where our sprite is horizontally
    ; as a result, the timing (CPU clocks) BEFORE the   sta RESPx
    ; in this scanline has to be accurate, to properly position sprite
    ; horizontally
    
    ldx #0                  ; 0 means no pixels, hiding sprite
    stx GRP0                ; set sprite 0 pixels
    
    ; wait for the beam to reach far enough to the right
    ; the sprite is displayed by the TIA on every single scan line
    ; but we will make sure to change the sprite's pixel data to
    ; "invisible" - that is, no pixels - on the scan lines where it
    ; shouldn't appear
    ldx SPRITE0_LEFT_X_COARSE
    inx                     ; we decrement immediately, so start one ahead
    inx                     ; ... and once more to account for a possible
                            ; -8 fine horizontal positioning
pad_x
    dex
    bne pad_x               ; busy wait loop, waiting for beam to reach
                            ; the sprite's horizontal location
    
    ; the beam has now reached the sprite's horizontal location
    sta RESP0               ; begin displaying sprite 0
    
    sta WSYNC               ; halt for scanline
    
    sta HMOVE               ; fine-adjust horizontal position of sprite 0
                            ; this is a good place to do it because:
                            ;     1. it is reached once per frame, and
                            ;     2. it is immediately after a WSYNC, as
                            ;        required by the spec
    
    ; scanline 36
    
    ; see if we have to advance animation frame
    inc FRAME_COUNTER
    lda #15
    bit FRAME_COUNTER
    bne after_advancing_animation_frame
    
    ; advance animation frame
    clc
    lda SPRITE0_PIXEL_OFFSET
    adc SPRITE_HEIGHT
    cmp SPRITE_TOTAL_FRAME_LINES    ; are we past the last frame?
    bne store_new_animation_frame   ; no
    ; yes, so go back to frame 0
    lda #0
store_new_animation_frame
    sta SPRITE0_PIXEL_OFFSET    ; move offset forward by sprite height
    
after_advancing_animation_frame
    lda #0
    sta CURRENT_SCANLINE    ; CURRENT_SCANLINE := 0
    
    ldy SPRITE0_PIXEL_OFFSET    ; Y := offset of first pixel line of frame
    
    sta WSYNC               ; halt for scanline
    
    ; scanline 37
    
    sta WSYNC               ; halt for scanline
    
    ; by halting immediately above, we ensure that we change the background
    ; colour immediately below, as the first visible begins (and while in
    ; the horizontal blanking interval)
    ; the reason for this is to change the background colour before the first
    ; pixel of the scanline is drawn
;=============================================================================
; 192 visible scanlines must now be drawn
;=============================================================================
    ldx #SKY_COLOUR
    stx COLUBK              ; set background colour
    
visible_scanlines_loop
; invariant: Y = offset of current line of pixels of sprite to display
; here, A = current scanline

    ; see if we have to turn on sprite 0
    cmp SPRITE0_BOTTOM_Y    ; is current scanline >= sprite bottom Y?
    bcs turn_off_sprite     ; yes
    cmp SPRITE0_TOP_Y       ; is current scanline < sprite top Y?
    bcc turn_off_sprite     ; yes
turn_on_sprite
    ; one of the sprite's horizontal lines must be displayed on this
    ; scanline
    ; NOTE: the sprite has already been "positioned" horizontally, thus
    ;       the logic here just writes the pixel data (1 byte) to the
    ;       pertinent register
    lda sprite_bitmap,y     ; A := current line of pixels of sprite 0
    sta GRP0                ; set sprite 0 pixels
    iny                     ; next line of pixels
    jmp after_sprite
turn_off_sprite
    lda #0
    sta GRP0                ; set sprite 0 pixels
after_sprite
    

visible_scanlines_next    
    inc CURRENT_SCANLINE
    lda CURRENT_SCANLINE
    cmp #192 - BOTTOM_LAYER_HEIGHT  ; this is how many scanlines we draw

    sta WSYNC               ; scanline
    ; MUST reach here with A = current scan line
    bne visible_scanlines_loop
    
    
; now draw the remaining scanlines at the bottom of the screen
; sprite cannot reach down here

    ldx #$B1                ; X := bottom background starting colour
bottom_scanlines
    cmp #192
    beq vertical_blanking
    
    ; change background colour to draw the bottom layer

    inx                     ;
    inx                     ;
    inx                     ;
    inx                     ; advance bottom background colour
    stx COLUBK              ; set background colour
    
    inc CURRENT_SCANLINE
    lda CURRENT_SCANLINE
    
    sta WSYNC               ; scanline
    ; MUST reach here with A = current scan line
    bne bottom_scanlines
    
vertical_blanking
    ; enter vertical blanking
    lda #%01000010
    sta VBLANK

;=============================================================================
; 30 scanlines of overscan
;=============================================================================
    REPEAT 30
        sta WSYNC
    REPEND
    
    jmp video_frame_start



    ; interrupt handlers
    ORG $FFFA
    .word start          ; NMI
    .word start          ; start
    .word start          ; IRQ
END
